/* -------------------------------------------------------------------------- */
/* $Id: gomp_support.c,v 1.1 2005/01/17 19:56:07 Jacob_Weismann_Poulsen Exp $ */
/* -------------------------------------------------------------------------- */

#define _ISOC9X_SOURCE 
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pthread.h>
#include <unistd.h>
#include <stdarg.h> 
#include <math.h>
#include "gomp_support.h"

#ifndef MIN
#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#endif

/* -------------------------------------------------------------------------- */
/* GLOBAL vars                                                                */
 
static pthread_t *my_threads; /* an array of thread ids */
static _OMP_state my_state;
static _OMP_loop_long_t *my_loopinfo;


/* -------------------------------------------------------------------------- */
/* GOMP trace BEGIN                                                           */

static unsigned int __gomp_debug_mask = 0;

void
TRACE (unsigned int module, unsigned int level, const char *format, ...)
{
  unsigned int mask = module << level;
  if (__gomp_debug_mask & mask)
    {
      char buf[255];		/* FIXME hardcoding */
      va_list ap;
      va_start (ap, format);
      vsnprintf (buf, sizeof (buf), format, ap);
      va_end (ap);
      printf ("TRACE %s", buf);
    }
}

void
SET_TRACE_LEVEL (unsigned int module, unsigned int level)
{
  unsigned int i;
  for (i = 0; i <= level; i++)
    __gomp_debug_mask |= module << i;
}

/* -------------------------------------------------------------------------- */
/* GOMP trace END                                                             */


/* -------------------------------------------------------------------------  */
/* fake wrapping of functions that belongs to libgomp */

/* FIXME */
int omp_get_num_threads(void) {
	return NUM_THREADS; /* FIXME fake wrapping of lib fct */
}

/* FIXME */
int omp_get_thread_num (void)
{
	int i = 0;
	int num_threads;
	pthread_t me;

	TRACE (GOMP_OMPLIB, 2, "BEGIN omp_get_thread_num\n");

	num_threads = omp_get_num_threads(); 
	me = pthread_self ();

	TRACE (GOMP_OMPLIB, 1, "thread me = %d\n", me);

	for (i = 0; i < num_threads; i++)
		if (pthread_equal (me, my_threads[i]))
			break;

	if (i ==num_threads)
   		_gomp_error(); 

	TRACE (GOMP_OMPLIB, 2, "END omp_get_thread_num\n");
	return i;
}

/* FIXME */
void _gomp_error(void) {
	printf("GOMP_error occured - FIXME");
	exit(1);
}

/* -------------------------------------------------------------------------  */
/* general support functions that belongs to libgomp */

/* FIXME */
void _OMP_init (void)
{
	int status;

	SET_TRACE_LEVEL (GOMP_SUPPORT, 0);
	SET_TRACE_LEVEL (GOMP_OMPLIB, 0);
	SET_TRACE_LEVEL (GOMP_LOOP, 2);

	my_threads = (pthread_t *)malloc(sizeof(pthread_t) * NUM_THREADS); 
	if (my_threads == NULL)
		_gomp_error(); 
	my_loopinfo = (_OMP_loop_long_t *)malloc(sizeof(_OMP_loop_long_t));
	if (my_loopinfo == NULL)
		_gomp_error(); 
	status = pthread_mutex_init (&my_loopinfo->mutex,NULL);
	if (status != 0)
		_gomp_error();  
	my_state.chunksize=DEFAULT_CHUNK;
	my_state.schedule=STATIC;
}

/* FIXME */
void _OMP_destroy (void)
{
	int status;
	free(my_threads); 
	status = pthread_mutex_destroy (&my_loopinfo->mutex);
	if (status != 0)
		_gomp_error();  
	free(my_loopinfo); 
}

/* FIXME */
void _OMP_PrivateVars(void *copyin_arg) {
	TRACE (GOMP_SUPPORT, 2, "_OMP_PrivateVars\n");
}

/* FIXME */
void _OMP_MasterFunction(_OMP_ParRegion_Info *master_thread, 
		void *arg, int par_cond, void *copyin_arg) {

   /* if par_cond [from if clause] is false then just do region serially */
   if (par_cond != 0) {
		/* wakeup slave threads; create threads if necessary */

		int n, status;
		int num_threads = omp_get_num_threads(); 

		/* an array of thread arguments */
		_OMP_Thread_Args *thread_args = 
			(_OMP_Thread_Args *)malloc(sizeof(_OMP_Thread_Args ) * num_threads);

		/* setup loop structures or do nothing if no loops in region */
		_OMP_LoopInit(master_thread);

		/* setup private vars structures for the threads */
		_OMP_PrivateVars(copyin_arg);

		for (n = 0; n < num_threads; ++n) {
			thread_args[n].thread_id = n;
			thread_args[n].args = arg;
			status = pthread_mutex_init(&thread_args[n].lock, NULL);
			if (status != 0)
				_gomp_error();
			status = pthread_mutex_lock(&thread_args[n].lock);
			if (status != 0)
				_gomp_error();
			status = pthread_create(
					 &my_threads[n],NULL,master_thread->func,&thread_args[n]);
			if (status != 0)
				_gomp_error();
		}
		for (n = 0; n < num_threads; ++n) {
			status = pthread_mutex_unlock(&thread_args[n].lock);
			if (status != 0)
				_gomp_error();
		}

		/* barrier -- wait for each thread in succession to finish */
		for (n = 0; n < num_threads; ++n) {
			status = pthread_join(my_threads[n],NULL);
			 if (status != 0)
				 _gomp_error();
		}

		/* clean up */
		free(thread_args);
	}
}

/* FIXME */
void _OMP_WorkSharing(_OMP_ParRegion_Info *o_pr_i, void *arg, int nowait) 
{
	TRACE (GOMP_SUPPORT, 2, "BEGIN _OMP_WorkSharing\n");
	TRACE (GOMP_SUPPORT, 0, "_OMP_WorkSharing FIXME\n");
	TRACE (GOMP_SUPPORT, 2, "END _OMP_WorkSharing\n");
}

/* -------------------------------------------------------------------------  */
/* loop related support functions that belongs to libgomp */

/* FIXME */
void _OMP_LoopInit_long(_OMP_ParRegion_Info *_pr_i) {
	/* Fill the loop_structure with correct values taken
	 * for _pr_i and the current state */
	my_loopinfo->done=0; 
	my_loopinfo->lower = _pr_i->lower;
	my_loopinfo->upper = _pr_i->upper;
	my_loopinfo->step = _pr_i->step;
	if (_pr_i->schedule == DEFAULT) {
		my_loopinfo->schedule = my_state.schedule;
		TRACE (GOMP_LOOP, 0, "Changing DEFAULT loop schedule = %i\n", my_loopinfo->schedule);
	}
	else {
		my_loopinfo->schedule = _pr_i->schedule;
		TRACE (GOMP_LOOP, 0, "NON DEFAULT loop schedule = %i\n", my_loopinfo->schedule);
	}
	if (_pr_i->chunksize == 0) {
		TRACE (GOMP_LOOP, 0, "loop chunk = current state %i\n", my_state.chunksize);
		my_loopinfo->chunk = my_state.chunksize; /* FIXME critical reg */
		_pr_i->chunksize = my_state.chunksize; /* FIXME critical reg */
	}
	else {
		TRACE (GOMP_LOOP, 0, "loop chunk = %i\n", _pr_i->chunksize);
		my_loopinfo->chunk = _pr_i->chunksize;
	}
}

void _OMP_LoopInit(_OMP_ParRegion_Info *_pr_i) {
	/* distinguish on type and call _OMP_LoopInit_<type> */
	TRACE (GOMP_LOOP, 2, "BEGIN _OMP_LoopInit\n");
	switch (_pr_i->indextype) {
		case INDEXTYPE_LONG:
			TRACE (GOMP_LOOP, 0, "_o_pr_i contains INDEXTYPE loog\n");
			_OMP_LoopInit_long(_pr_i);
			break;
		case INDEXTYPE_SHORT:
/*			_OMP_LoopInit_short(_pr_i);*/
			break;
		case INDEXTYPE_INT:
/*			_OMP_LoopInit_int(_pr_i);*/
			break;
		default:
			TRACE (GOMP_LOOP, 0, "_o_pr_i contains no INDEXTYPE, i.e. not loop\n");
			break;
		}
	TRACE (GOMP_LOOP, 2, "END _OMP_LoopInit\n");

}

long _OMP_GetNextChunk_long(long *my_lower, long *my_upper) {
	/* returns 0: have another chunk; 1: no more chunks */
	int status, iam, nthreads;
	long loopsize, chunk, rest;

	TRACE (GOMP_LOOP, 2, "BEGIN _OMP_GetNextChunk_long\n");
	iam = omp_get_thread_num();
	nthreads = omp_get_num_threads();
	switch (my_loopinfo->schedule) {
		case STATIC:
			TRACE (GOMP_LOOP, 1, "state is static\n");
			status = pthread_mutex_lock (&my_loopinfo->mutex);
			if (status != 0)
				 _gomp_error(); 
			loopsize = my_loopinfo->upper - my_loopinfo->lower+1;
			chunk = lrint(loopsize/(long)nthreads);
			rest = loopsize - nthreads*chunk;
			TRACE (GOMP_LOOP, 1, "lower is %li, upper is %li, loopsize is %li, chunk is %li, rest is %li\n", my_loopinfo->lower, my_loopinfo->upper, loopsize, chunk, rest);
			if (iam != 0)
				*my_lower = my_loopinfo->lower+iam*chunk+rest;
			else {
				chunk = chunk+rest;
				*my_lower = my_loopinfo->lower;
			}
			*my_upper = *my_lower+chunk-1;
			TRACE (GOMP_LOOP, 1, "my_id %i, my_lower %li, my_upper %li\n", 
					iam,*my_lower,*my_upper);
			my_loopinfo->done=1; 
			status = pthread_mutex_unlock (&my_loopinfo->mutex);
			if (status != 0)
				 _gomp_error(); 
			break;
		case DYNAMIC:
			TRACE (GOMP_LOOP, 1, "state is dynamic\n");
			status = pthread_mutex_lock (&my_loopinfo->mutex);
			if (status != 0)
				 _gomp_error(); 
			TRACE (GOMP_LOOP, 1, "thread %i: my_loopinfo->lower before: %li\n", iam, my_loopinfo->lower);

			*my_lower = iam*my_loopinfo->chunk+my_loopinfo->lower;
			*my_upper = MIN((iam+1)*my_loopinfo->chunk+*my_lower-1, my_loopinfo->upper);
			my_loopinfo->lower = *my_upper+1;
			if (iam==0) {
				if (my_loopinfo->lower >= my_loopinfo->upper)
					my_loopinfo->done=1; 
			}
			else {
				if (my_loopinfo->lower + my_loopinfo->chunk >= my_loopinfo->upper)
					my_loopinfo->done=1;
			}
			TRACE (GOMP_LOOP, 1, "thread %i: my_loopinfo->lower after: %li\n", iam, my_loopinfo->lower);
			TRACE (GOMP_LOOP, 1, "my_lower %li, my_upper %li\n", 
					*my_lower,*my_upper);
			status = pthread_mutex_unlock (&my_loopinfo->mutex);
			if (status != 0)
				 _gomp_error(); 
			break;
		case GUIDED:
			TRACE (GOMP_LOOP, 1, "state is guided\n");
			break;
		case RUNTIME:
			TRACE (GOMP_LOOP, 1, "state is runtime\n");
			break;
		default:
			TRACE (GOMP_LOOP, 1, "state is default case\n");
			break;
		}
	TRACE (GOMP_LOOP, 2, "END _OMP_GetNextChunk_long\n");
	return my_loopinfo->done;
}

/* -------------------------------------------------------------------------  */

