/* -------------------------------------------------------------------------- */
/* $Id: main.c,v 1.1 2005/01/17 19:56:07 Jacob_Weismann_Poulsen Exp $ */
/* -------------------------------------------------------------------------- */

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pthread.h>
#include <stdarg.h>
#include "gomp_support.h"

void sample0(void);
void foo(void);
void sample1(float *a, float *b, float *c, long n);
void sample2(float *a, float *b, float *c, long n);
void sample3(float *a, float *b, float *c, long n);

/* -------------------------------------------------------------------------  */
/* sample0 - a single parallel region */
/* -------------------------------------------------------------------------  */

/*
void sample0() {
#pragma omp parallel
	printf("Hello, World from %i\n",omp_get_thread_num());
}
*/

static void *sample0_region0 (void *args) {
	_OMP_Thread_Args *ta;
	ta = (_OMP_Thread_Args *) args;

	pthread_mutex_lock(&ta->lock);

	TRACE (SAMPLE0, 2, "BEGIN\n");
	printf("Hello, World from %i\n",omp_get_thread_num());
	TRACE (SAMPLE0, 2, "END\n");
	pthread_exit(0);
	return NULL;
}

void sample0(void) {
	_OMP_ParRegion_Info _o_pr_i;
	_o_pr_i.func = &sample0_region0;
	_OMP_MasterFunction(&_o_pr_i,NULL, 1, NULL);
}

/* -------------------------------------------------------------------------  */
/* sample1 - two parallel worksharing regions. */
/* -------------------------------------------------------------------------  */

/*
extern void foo();

void sample(float *a, float *b, float *c, long n) {
  long i;

#pragma omp parallel
  foo();

#pragma omp parallel for
  for (i=0; i < n; i++)
    *a++ = *b++ + *c++;

}
*/

void foo(void) { return; }

struct _sample1_args {
	float *a;
	float *b;
	float *c;
	long n;
};

static void *sample1_region0(void *args) {
	_OMP_Thread_Args *ta;
	int thread_id;
	ta = (_OMP_Thread_Args *) args;
	thread_id = ta->thread_id;

	pthread_mutex_lock(&ta->lock);

	TRACE(SAMPLE1, 2, "region 0 BEGIN\n");
	TRACE(SAMPLE1, 1, "sample1 region0, thread id %i\n",thread_id);
	foo();
	TRACE(SAMPLE1, 2, "region 0 END\n");
	pthread_exit(0);
	return NULL;
}

static void *sample1_region1(void *args) {
	_OMP_Thread_Args *ta;
	int thread_id;
	struct _sample1_args *ra;
	float *a;
	float *b;
	float *c;
	long _chunk_lower, _chunk_upper, i;

	ta = (_OMP_Thread_Args *) args;
	thread_id = ta->thread_id;
	ra = (struct _sample1_args *) ta->args;
	a = ra->a;
	b = ra->b;
	c= ra->c;

	pthread_mutex_lock(&ta->lock);

	TRACE(SAMPLE1, 2, "region 1 BEGIN\n");

	TRACE(SAMPLE1, 1, "sample1 region1, thread id %i\n",thread_id);
	TRACE(SAMPLE1, 1, "float in sample1_region1 is %f\n",*a++);
	TRACE(SAMPLE1, 1, "float in sample1_region1 is %f\n",*a++);
	
	while (_OMP_GetNextChunk_long(&_chunk_lower, &_chunk_upper) != 1) {
		for (i = _chunk_lower; i <= _chunk_upper; i++)
			*a++ = *b++ + *c++;
	}
	TRACE(SAMPLE1, 2, "region 1 END\n");
	pthread_exit(0);
	return NULL;
}

void sample1(float *a, float *b, float *c, long n) {
	_OMP_ParRegion_Info _o_pr_i;

	_o_pr_i.func = &sample1_region0;
	_OMP_MasterFunction(&_o_pr_i, NULL, 1, NULL); 

	if (0 < n) {
		struct _sample1_args _args;
		_o_pr_i.func = &sample1_region1;
		_o_pr_i.lower = 0;
		_o_pr_i.upper = n-1;
		_o_pr_i.step = 1;
		_o_pr_i.indextype = INDEXTYPE_LONG;
		_o_pr_i.schedule = DEFAULT;
		_o_pr_i.chunksize = 0;
		_args.a = a;
		_args.b = b;
		_args.c = c;
		_args.n = n; 
		_OMP_MasterFunction(&_o_pr_i, (void *) &_args, 1, NULL); 
	}
}

/* -------------------------------------------------------------------------  */
/* sample2 - one parallel worksharing regions with both ordinary code and 
 *           a loop */
/* -------------------------------------------------------------------------  */

/*
 
extern void foo();

void sample(float *a, float *b, float *c, long n) {
  long i;

#pragma omp parallel
 { 
  foo();

#pragma omp for
  for (i=0; i < n; i++)
    *a++ = *b++ + *c++;

 } 
}

*/

struct _sample2_args {
	float *a;
	float *b;
	float *c;
	long n;
};

static void *sample2_region1(void *args) {
	_OMP_Thread_Args *ta;
	int thread_id;
	struct _sample2_args *ra;
	float *a;
	float *b;
	float *c;
	long _chunk_lower, _chunk_upper, i;

	ta = (_OMP_Thread_Args *) args;
	thread_id = ta->thread_id;
	ra = (struct _sample2_args *) ta->args;
	a = ra->a;
	b = ra->b;
	c= ra->c;

	pthread_mutex_lock(&ta->lock);

	TRACE(SAMPLE2, 2, "region 1 BEGIN\n");

	TRACE(SAMPLE2, 1, "sample2 region1, thread id %i\n",thread_id);
	TRACE(SAMPLE2, 1, "float in sample2_region1 is %f\n",*a++);
	TRACE(SAMPLE2, 1, "float in sample2_region1 is %f\n",*a++);

	while (_OMP_GetNextChunk_long(&_chunk_lower, &_chunk_upper) != 1) {
		for (i = _chunk_lower; i <= _chunk_upper; i++)
			*a++ = *b++ + *c++;
	}
	TRACE(SAMPLE2, 2, "region 1 END\n");
	pthread_exit(0);
	return NULL;
}

static void *sample2_region0(void *args) {
	_OMP_ParRegion_Info _o_pr_i;
	_OMP_Thread_Args *ta;
	int thread_id;
	struct _sample2_args *ra;
	float *a;
	float *b;
	float *c;
	long n;

	ta = (_OMP_Thread_Args *) args;
	thread_id = ta->thread_id;
	ra = (struct _sample2_args *) ta->args;
	a = ra->a;
	b = ra->b;
	c= ra->c;
	n= ra->n;

	TRACE(SAMPLE2, 1, "sample2 region0, thread id %i\n",thread_id);

	foo();

	if (0 < n) {
		struct _sample2_args _args;
		_o_pr_i.func = &sample2_region1;
		_o_pr_i.lower = 0;
		_o_pr_i.upper = n-1;
		_o_pr_i.step = 1;
		_o_pr_i.indextype = INDEXTYPE_LONG;
		_o_pr_i.schedule = DEFAULT;
		_o_pr_i.chunksize = 0;
		_args.a = a;
		_args.b = b;
		_args.c = c;
		_args.n = n; 
		_OMP_WorkSharing(&_o_pr_i, (void *) &_args, 0); 
	}
	pthread_exit(0);
	return NULL;
}

void sample2(float *a, float *b, float *c, long n) {
	_OMP_ParRegion_Info _o_pr_i;
	struct _sample2_args _args;

	_o_pr_i.func = &sample2_region0;
	_args.a = a;
	_args.b = b;
	_args.c = c;
	_args.n = n; 

	_OMP_MasterFunction(&_o_pr_i, (void *) &_args, 1, NULL); 
}

/* -------------------------------------------------------------------------  */
/* sample3 - loop with another schedule */
/* -------------------------------------------------------------------------  */

/*

void sample3(float *a, float *b, float *c, long n) {
  long i;

#pragma omp parallel for chunksize=8 schedule=dynamic 
		for (i=0; i < n; i++)
			*a++ = *b++ + *c++;
	} 
}

*/

struct _sample3_args {
	float *a;
	float *b;
	float *c;
	long n;
};

static void *sample3_region0(void *args) {
	_OMP_Thread_Args *ta;
	int thread_id;
	struct _sample3_args *ra;
	float *a;
	float *b;
	float *c;
	long _chunk_lower, _chunk_upper, i;

	ta = (_OMP_Thread_Args *) args;
	thread_id = ta->thread_id;
	ra = (struct _sample3_args *) ta->args;
	a = ra->a;
	b = ra->b;
	c= ra->c;

	pthread_mutex_lock(&ta->lock);

	TRACE (SAMPLE3, 2, "reg0 BEGIN\n");

	TRACE(SAMPLE3, 1, "sample3 region0, thread id %i\n",thread_id);
	TRACE(SAMPLE3, 1, "float in sample3_region0 is %f\n",*a++);
	TRACE(SAMPLE3, 1, "float in sample3_region0 is %f\n",*a++);
	
	while (_OMP_GetNextChunk_long(&_chunk_lower, &_chunk_upper) != 1) {
		for (i = _chunk_lower; i <= _chunk_upper; i++)
			*a++ = *b++ + *c++;
	}
	TRACE (SAMPLE3, 2, "reg0 END\n");
	pthread_exit(0);
	return NULL;
}

void sample3(float *a, float *b, float *c, long n) {
	_OMP_ParRegion_Info _o_pr_i;

	if (0 < n) {
		struct _sample3_args _args;
		_o_pr_i.func = &sample3_region0;
		_o_pr_i.lower = 0;
		_o_pr_i.upper = n-1;
		_o_pr_i.step = 1;
		_o_pr_i.indextype = INDEXTYPE_LONG;
		_o_pr_i.schedule = DYNAMIC;
		_o_pr_i.chunksize = 8;
		_args.a = a;
		_args.b = b;
		_args.c = c;
		_args.n = n; 
		_OMP_MasterFunction(&_o_pr_i, (void *) &_args, 1, NULL); 
	}
}

/* -------------------------------------------------------------------------  */


/* -------------------------------------------------------------------------  */

#define VECTOR_SIZE 100000
/* #define VECTOR_SIZE 8 */

int main(void) {
	float a[VECTOR_SIZE];
	float b[VECTOR_SIZE];
	float c[VECTOR_SIZE];

	SET_TRACE_LEVEL (SAMPLE0, 0);
	SET_TRACE_LEVEL (SAMPLE1, 0);
	SET_TRACE_LEVEL (SAMPLE2, 0);
	SET_TRACE_LEVEL (SAMPLE3, 2);

	_OMP_init();

	printf("-- hello_word sample --\n\n");
	sample0();

	printf("\n\n-- loop sample --\n\n");
	a[0]=1.0;
	a[1]=4.0;
	sample1(&a[0],&b[0],&c[0],VECTOR_SIZE);

	printf("\n\n-- parallel sample --\n\n");
	a[0]=3.0;
	a[1]=2.0;
	sample2(&a[0],&b[0],&c[0],VECTOR_SIZE);

	printf("\n\n-- loop sample - dynamic schedule --\n\n");
	sample3(&a[0],&b[0],&c[0],VECTOR_SIZE);

	_OMP_destroy();

	return 0;
}

/* -------------------------------------------------------------------------  */

